# Copyright 2017 Klarna Bank AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.conf import settings
from django.contrib.auth.models import Group
from rest_framework import serializers
import netaddr
import re
from django_rethink import r, RethinkSerializer, RethinkObjectNotFound, RethinkMultipleObjectsFound, validate_unique_key, get_connection, HistorySerializerMixin, NeedsReviewMixin, PermissionsSerializer
from django_rethink.tasks import rethinkdb_lock, rethinkdb_unlock

def validate_group_name(group_name):
    try:
        group = Group.objects.get(name=group_name)
        return True
    except Group.DoesNotExist:
        if hasattr(settings, 'AUTH_LDAP_SERVER_URI'):
            import ldap
            l = ldap.initialize(settings.AUTH_LDAP_SERVER_URI)
            if settings.AUTH_LDAP_START_TLS:
                l.start_tls_s()
            result = settings.AUTH_LDAP_GROUP_SEARCH.search_with_additional_term_string("(cn=%s)").execute(l, filterargs=(group_name,))
            if len(result) > 0:
                return True
        raise serializers.ValidationError("group %s does not exist" % group_name)

class BonkTriggerMixin(object):
    def create(self, data):
        import bonk.tasks
        data = super(BonkTriggerMixin, self).create(data)
        task = rethinkdb_lock.s(name='trigger_dns_dhcp_rebuild') | \
            bonk.tasks.trigger_dns_dhcp_rebuild.si(data) | \
            rethinkdb_unlock.si(name='trigger_dns_dhcp_rebuild')
        task.apply_async()
        return data

    def update(self, instance, data):
        import bonk.tasks
        data = super(BonkTriggerMixin, self).update(instance, data)
        task = rethinkdb_lock.s(name='trigger_dns_dhcp_rebuild') | \
            bonk.tasks.trigger_dns_dhcp_rebuild.si(data) | \
            rethinkdb_unlock.si(name='trigger_dns_dhcp_rebuild')
        task.apply_async()
        return data

    def delete(self):
        import bonk.tasks
        ret = super(BonkTriggerMixin, self).delete()
        task = rethinkdb_lock.s(name='trigger_dns_dhcp_rebuild') | \
            bonk.tasks.trigger_dns_dhcp_rebuild.si(self.instance) | \
            rethinkdb_unlock.si(name='trigger_dns_dhcp_rebuild')
        task.apply_async()
        return ret

class VRFSerializer(HistorySerializerMixin):
    id = serializers.CharField(required=False)
    tags = serializers.DictField(required=False)
    vrf = serializers.IntegerField(required=True)
    name = serializers.CharField(required=True)

    class Meta(RethinkSerializer.Meta):
        table_name = 'vrf'
        slug_field = 'vrf'
        indices = [
            'vrf',
        ]

def validate_vrf(value):
    try:
        VRFSerializer.get(vrf=value)
    except:
        raise serializers.ValidationError("vrf=%r doesn't exist" % value)
    return value

class IPBlockSerializer(HistorySerializerMixin):
    id = serializers.CharField(required=False)
    tags = serializers.DictField(required=False)
    vrf = serializers.IntegerField(required=True, validators=[validate_vrf])
    network = serializers.IPAddressField(required=True)
    length = serializers.IntegerField(required=True)
    announced_by = serializers.CharField(required=False)
    permissions = PermissionsSerializer(required=False)

    class Meta(RethinkSerializer.Meta):
        table_name = 'ip_block'
        slug_field = 'vrf_network_length'
        indices = [
            ('vrf_network_length', (r.row['vrf'], r.row['network'], r.row['length'])),
            ('permissions_read', r.row['permissions']['read'], {'multi': True}),
            ('permissions_create', r.row['permissions']['create'], {'multi': True}),
            ('permissions_write', r.row['permissions']['write'], {'multi': True}),
        ]
        unique_together = [
            ('vrf', 'network', 'length'),
        ]

    @classmethod
    def get_by_ip(cls, vrf, ip, reql=False):
        query = cls.filter(lambda b:
                    r.ip_prefix_contains(
                        r.ip_prefix(b['network'], b['length']),
                        r.ip_address(ip)
                    ), reql=True) \
                .filter({'vrf': vrf}) \
                .order_by(r.desc("length")).nth(0)
        if reql:
            return query
        else:
            try:
                return query.run(get_connection())
            except r.errors.ReqlNonExistenceError:
                raise RethinkObjectNotFound("no block found for IP %s" % ip)

    @classmethod
    def filter_by_block(cls, block, reql=False):
        return cls.filter(lambda b:
            r.ip_prefix_contains(
                r.ip_prefix(block['network'], block['length']),
                r.ip_address(b['network'])
            ), reql=reql)

    def validate(self, data):
        data = super(IPBlockSerializer, self).validate(data)
        full = self.get_updated_object(data)
        network = netaddr.IPNetwork("%s/%d" % (full['network'], full['length']))
        if str(network.network) != full['network']:
            raise serializers.ValidationError("network is not the network address for %s/%d" % (full['network'], full['length']))
        return data

class IPPrefixDHCPSerializer(serializers.Serializer):
    enabled = serializers.BooleanField(required=True)
    server_set = serializers.CharField(required=False)
    range = serializers.ListField(child=serializers.IPAddressField(), required=False)
    options = serializers.ListField(child=serializers.CharField(), required=False)

    def validate_range(self, value):
        if value and len(value) != 2:
            raise serializers.ValidationError("range must have a start and end address")
        return value

    def validate_server_set(self, value):
        try:
            DHCPServerSetSerializer.get(name=value)
        except RethinkObjectNotFound:
            raise serializers.ValidationError("server_set=%r doesn't exist" % value)
        return value

class DDNSSerializer(serializers.Serializer):
    name = serializers.CharField(required=True)
    algorithm = serializers.CharField(required=True)
    key = serializers.CharField(required=True)

class IPPrefixDDNSSerializer(DDNSSerializer):
    zone = serializers.CharField(required=True)
    server = serializers.IPAddressField(required=True)

class IPPrefixSerializer(BonkTriggerMixin, HistorySerializerMixin):
    id = serializers.CharField(required=False)
    tags = serializers.DictField(required=False)
    vrf = serializers.IntegerField(required=True, validators=[validate_vrf])
    network = serializers.IPAddressField(required=True)
    length = serializers.IntegerField(required=True)
    asn = serializers.IntegerField(required=False)
    name = serializers.CharField(required=False)
    state = serializers.ChoiceField(required=True, choices=['allocated', 'reserved', 'quarantine'])
    permissions = PermissionsSerializer(required=False)
    gateway = serializers.IPAddressField(required=False)
    dhcp = IPPrefixDHCPSerializer(required=False)
    ddns = IPPrefixDDNSSerializer(required=False)
    reference = serializers.CharField(required=False)
    inventory_id = serializers.CharField(required=False)

    class Meta(RethinkSerializer.Meta):
        table_name = 'ip_prefix'
        slug_field = 'vrf_network_length'
        indices = [
            'name',
            ('vrf_network_length', (r.row['vrf'], r.row['network'], r.row['length'])),
            ('permissions_read', r.row['permissions']['read'], {'multi': True}),
            ('permissions_create', r.row['permissions']['create'], {'multi': True}),
            ('permissions_write', r.row['permissions']['write'], {'multi': True}),
        ]
        unique = [
            'name'
        ]
        unique_together = [
            ('vrf', 'network', 'length'),
        ]

    @classmethod
    def filter_by_block(cls, block, reql=False):
        return cls.filter(lambda p:
            r.ip_prefix_contains(
                r.ip_prefix(block['network'], block['length']),
                r.ip_address(p['network'])
            ), reql=reql)

    @classmethod
    def get_by_ip(cls, vrf, ip, reql=False):
        query = cls.filter(lambda p:
                r.ip_prefix_contains(
                    r.ip_prefix(p['network'], p['length']),
                    r.ip_address(ip)
                ), reql=True) \
                .filter({'vrf': vrf}) \
                .order_by(r.desc("length")).nth(0)
        if reql:
            return query
        else:
            try:
                return query.run(get_connection())
            except r.errors.ReqlNonExistenceError:
                raise RethinkObjectNotFound("no prefix found for IP %s" % ip)

    def validate(self, data):
        data = super(IPPrefixSerializer, self).validate(data)
        full = self.get_updated_object(data)
        try:
            block = IPBlockSerializer.get_by_ip(full['vrf'], full['network'])
        except RethinkObjectNotFound:
            raise serializers.ValidationError("no block exists matching prefix %s/%d" % (full['network'], full['length']))
        if block['length'] > full['length']:
            raise serializers.ValidationError("prefix %s/%d exceeds block of %s/%d" % (full['network'], full['length'], block['network'], block['length']))
        if (self.instance is None and
                self.context['request'].user is not None and
                not self.context['request'].user.is_superuser
            ):
            allowed = set(
                block.get('permissions', {}).get('write', []) +
                block.get('permissions', {}).get('create', [])
            )
            groups = set(self.context['request'].user.groups.all().values_list('name', flat=True))
            if len(groups.intersection(allowed)) == 0:
                raise serializers.ValidationError("you do not have permissions to block %s/%d" % (block['network'], block['length']))
        underlappers = filter(lambda x: x[self.Meta.pk_field] != full.get(self.Meta.pk_field, None), self.filter_by_block(full))
        if len(underlappers) > 0:
            raise serializers.ValidationError("prefix %s/%d overlaps with %r" % (full['network'], full['length'], underlappers))
        try:
            overlapper = IPPrefixSerializer.get_by_ip(full['vrf'], full['network'])
            if overlapper[self.Meta.pk_field] != full.get(self.Meta.pk_field, None):
                raise serializers.ValidationError("prefix %s/%d includes this prefix %s/%d" % (overlapper['network'], overlapper['length'], full['network'], full['length']))
        except RethinkObjectNotFound:
            pass
        network = netaddr.IPNetwork("%s/%d" % (full['network'], full['length']))
        if str(network.network) != full['network']:
            raise serializers.ValidationError("network is not the network address for %s/%d" % (full['network'], full['length']))
        return data

    def create(self, data):
        import bonk.tasks
        data = super(IPPrefixSerializer, self).create(data)
        block = IPBlockSerializer.get_by_ip(data['vrf'], data['network'])
        if 'announced_by' in block and data['state'] == 'allocated':
            bonk.tasks.trigger_prefix_create.apply_async((data, block))
        return data

    def update(self, instance, data):
        import bonk.tasks
        ret = super(IPPrefixSerializer, self).update(instance, data)
        block = IPBlockSerializer.get_by_ip(ret['vrf'], ret['network'])
        if 'announced_by' in block:
            if instance['state'] != 'allocated' and ret['state'] == 'allocated':
                bonk.tasks.trigger_prefix_create.apply_async((ret, block))
            elif instance['state'] == 'allocated' and ret['state'] != 'allocated':
                bonk.tasks.trigger_prefix_delete.apply_async((ret, block))
        return ret

    def delete(self):
        import bonk.tasks
        block = IPBlockSerializer.get_by_ip(self.instance['vrf'], self.instance['network'])
        if 'announced_by' in block and self.instance['state'] == 'allocated':
            bonk.tasks.trigger_prefix_delete.apply_async((self.instance, block))
        ret = super(IPPrefixSerializer, self).delete()
        return ret

validate_mac_re = re.compile(r'^(?:[0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$')
def validate_mac(value):
    if validate_mac_re.match(value) is None:
        raise serializers.ValidationError("%s is not a valid MAC address (format as de:ad:be:ef:f0:00)" % value)

validate_fqdn_re = re.compile(r'^(([a-zA-Z0-9_][a-zA-Z0-9\-]*[a-zA-Z0-9]|[a-zA-Z0-9]|\*)\.)*([A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$')
def validate_fqdn(value):
    if validate_fqdn_re.match(value) is None:
        raise serializers.ValidationError("%s is not a valid FQDN" % value)

def validate_ttl(value):
    if value.bit_length() > 32:
        raise serializers.ValidationError("TTL can't be larger than 32 bits, value: %s" % value)
    if value < 0:
        raise serializers.ValidationError("TTL must be a positive value, value: %s")

class IPAddressSerializer(BonkTriggerMixin, HistorySerializerMixin):
    id = serializers.CharField(required=False)
    tags = serializers.DictField(required=False)
    state = serializers.ChoiceField(required=True, choices=['allocated', 'reserved', 'quarantine'])
    vrf = serializers.IntegerField(required=True, validators=[validate_vrf])
    ip = serializers.IPAddressField(required=True)
    name = serializers.CharField(required=True, validators=[validate_fqdn])
    dhcp_mac = serializers.ListField(child=serializers.CharField(validators=[validate_mac]), required=False)
    reference = serializers.CharField(required=False)
    permissions = PermissionsSerializer(required=False)
    ttl = serializers.IntegerField(required=False, validators=[validate_ttl])

    class Meta(RethinkSerializer.Meta):
        table_name = 'ip_address'
        slug_field = 'vrf_ip'
        indices = [
            'ip',
            'name',
            ('vrf_ip', (r.row['vrf'], r.row['ip'])),
            ('permissions_read', r.row['permissions']['read'], {'multi': True}),
            ('permissions_create', r.row['permissions']['create'], {'multi': True}),
            ('permissions_write', r.row['permissions']['write'], {'multi': True}),
        ]
        unique_together = [
            ('vrf', 'ip'),
        ]

    @classmethod
    def filter_by_prefix(cls, prefix, reql=False):
        return cls.filter(lambda a:
            r.ip_prefix_contains(
                r.ip_prefix(prefix['network'], prefix['length']),
                r.ip_address(a['ip'])
            ),
        reql=reql)

    def validate_name(self, value):
        possibles = []
        for part in value.split(".")[:0:-1]:
            suffix = "" if len(possibles) == 0 else ("." + possibles[-1])
            possibles.append(part + suffix)
        try:
            zone = DNSZoneSerializer.filter(lambda zone: r.expr(possibles).contains(zone['name']), reql=True).order_by(r.desc(r.row['name'].count())).nth(0).run(self.conn)
        except r.errors.ReqlNonExistenceError:
            raise serializers.ValidationError("no zone matching %s could be found" % value)
        if 'request' in self.context and not self.context['request'].user.is_superuser:
            user_groups = set(self.context['request'].user.groups.all().values_list('name', flat=True))
            if self.instance is not None and len(user_groups.intersection(set(
                    self.instance.get('permissions', {}).get('write', [])
                ))) > 0:
                pass
            elif len(user_groups.intersection(set(
                    zone.get('permissions', {}).get('create', []) +
                    zone.get('permissions', {}).get('write', [])
                ))) == 0:
                raise serializers.ValidationError("you do not have permission to create names in %s" % zone['name'])
        try:
            ip_address = IPAddressSerializer.get(name=value)
            if self.instance is None or ip_address['id'] != self.instance['id']:
                raise serializers.ValidationError("%r is already in use by %s" % (value, ip_address['ip']))
        except RethinkObjectNotFound:
            pass
        return value

    def validate(self, data):
        data = super(IPAddressSerializer, self).validate(data)
        full = self.get_updated_object(data)
        try:
            prefix = IPPrefixSerializer.get_by_ip(full['vrf'], full['ip'])
        except RethinkObjectNotFound:
            raise serializers.ValidationError("no prefix found for IP %s" % full['ip'])
        return data

class DNSZoneOptionsSerializer(serializers.Serializer):
    ddns = DDNSSerializer(required=False)
    forwarders = serializers.ListField(child=serializers.IPAddressField(), required=False)
    notify = serializers.ListField(child=serializers.IPAddressField(), required=False)
    masters = serializers.ListField(child=serializers.IPAddressField(), required=False)

class DNSSOASerializer(serializers.Serializer):
    authns = serializers.CharField(required=True)
    email = serializers.CharField(required=True)
    refresh = serializers.IntegerField(required=True)
    retry = serializers.IntegerField(required=True)
    expiry = serializers.IntegerField(required=True)
    nxdomain = serializers.IntegerField(required=True)

class DNSZoneSerializer(NeedsReviewMixin, BonkTriggerMixin, HistorySerializerMixin):
    id = serializers.CharField(required=False)
    tags = serializers.DictField(required=False)
    needs_review = serializers.BooleanField(required=False, default=False)
    type = serializers.ChoiceField(required=True, choices=['internal', 'external'])
    name = serializers.CharField(required=True, validators=[validate_fqdn])
    soa = DNSSOASerializer(required=False)
    ttl = serializers.IntegerField(required=False, validators=[validate_ttl])
    options = DNSZoneOptionsSerializer(required=False)
    permissions = PermissionsSerializer(required=False)

    class Meta(RethinkSerializer.Meta):
        table_name = 'dns_zone'
        slug_field = 'name'
        needs_review_field = 'needs_review'
        indices = [
            'name',
            ('permissions_read', r.row['permissions']['read'], {'multi': True}),
            ('permissions_create', r.row['permissions']['create'], {'multi': True}),
            ('permissions_write', r.row['permissions']['write'], {'multi': True}),
        ]
        unique = [
            'name',
        ]

    def validate(self, data):
        data = super(DNSZoneSerializer, self).validate(data)
        if (self.instance is None and
                self.context['request'].user is not None and
                not self.context['request'].user.is_superuser
            ):
            zones = DNSZoneSerializer.filter(type=data['type'])
            parent = {'name': '.'}
            for zone in zones:
                if data['name'].endswith("." + zone['name']):
                    if len(zone['name']) > len(parent['name']):
                        parent = zone
            allowed = set(
                parent.get('permissions', {}).get('write', []) +
                parent.get('permissions', {}).get('create', [])
            )
            groups = set(self.context['request'].user.groups.all().values_list('name', flat=True))
            if len(groups.intersection(allowed)) == 0:
                raise serializers.ValidationError("you do not have permissions to zone %s" % (parent['name']))
        return data

class DNSRecordSerializer(NeedsReviewMixin, BonkTriggerMixin, HistorySerializerMixin):
    id = serializers.CharField(required=False)
    name = serializers.CharField(required=True, validators=[validate_fqdn])
    zone = serializers.CharField(required=True)
    type = serializers.ChoiceField(choices=['A', 'AAAA', 'CNAME', 'MX', 'NS', 'PTR', 'SRV', 'TXT'], required=True)
    ttl = serializers.IntegerField(required=False, validators=[validate_ttl])
    value = serializers.ListField(child=serializers.CharField())
    reference = serializers.CharField(required=False)
    permissions = PermissionsSerializer(required=False)

    class Meta(RethinkSerializer.Meta):
        table_name = 'dns_record'
        slug_field = 'name_type'
        indices = [
            'name',
            'zone',
            ('name_type', (r.row['name'], r.row['type'])),
            ('zone_name_type', (r.row['zone'], r.row['name'], r.row['type'])),
            ('value', {'multi': True}),
            ('permissions_read', r.row['permissions']['read'], {'multi': True}),
            ('permissions_create', r.row['permissions']['create'], {'multi': True}),
            ('permissions_write', r.row['permissions']['write'], {'multi': True}),
        ]
        unique_together = [
            ('zone', 'name', 'type'),
        ]

    def needs_review(self, instance, data):
        if not hasattr(self, 'zone'):
            self._zone = DNSZoneSerializer(DNSZoneSerializer.get(name=data.get('zone', instance['zone'] if instance else None)))
        return self._zone.needs_review(self._zone.instance, {})

    def get_reviewers(self, instance, data):
        return self._zone.get_reviewers(self._zone.instance, {})

    def validate_zone(self, value):
        try:
            self._zone = DNSZoneSerializer(DNSZoneSerializer.get(name=value))
        except RethinkObjectNotFound:
            raise serializers.ValidationError("'%s' does not exist" % value)
        if 'request' in self.context and not self.context['request'].user.is_superuser:
            user_groups = set(self.context['request'].user.groups.all().values_list('name', flat=True))
            if self.instance is not None and len(user_groups.intersection(set(
                    self.instance.get('permissions', {}).get('write', [])
                ))) > 0:
                pass
            elif len(user_groups.intersection(set(
                    self._zone.instance.get('permissions', {}).get('create', []) +
                    self._zone.instance.get('permissions', {}).get('write', [])
                ))) == 0:
                raise serializers.ValidationError("you do not have permission to create names in %s" % value)
        return value

    def validate(self, data):
        data = super(DNSRecordSerializer, self).validate(data)
        full = self.get_updated_object(data)
        if full['name'] != full['zone'] and not full['name'].endswith('.' + full['zone']):
            raise serializers.ValidationError("name %s is not in zone %s" % (full['name'], full['zone']))
        # FIXME: Add validation of value for type
        if full['type'] == 'CNAME':
            records = list(DNSRecordSerializer.filter(name=full['name']))
            if self.instance is not None:
                records = filter(lambda x: x['id'] != self.instance['id'], records)
            if len(records) > 0:
                raise serializers.ValidationError("a CNAME record cannot be used on a name with any other record type")
            if len(full['value']) > 1:
                raise serializers.ValidationError("a CNAME record can only have one value")
        else:
            records = list(DNSRecordSerializer.filter(name=full['name'], type='CNAME'))
            if self.instance is not None:
                records = filter(lambda x: x['id'] != self.instance['id'], records)
            if len(records) > 0:
                raise serializers.ValidationError("a CNAME record exists for the specified name already")
        return data

class DHCPServerSetSerializer(BonkTriggerMixin, HistorySerializerMixin):
    id = serializers.CharField(required=False)
    name = serializers.CharField(required=True)
    servers = serializers.ListField(child=serializers.IPAddressField())

    class Meta(RethinkSerializer.Meta):
        table_name = 'dhcp_server_set'
        slug_field = 'name'
        indices = [
            'name',
        ]
        unique = [
            'name',
        ]



# Copyright 2017 Klarna Bank AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import os
import netaddr
from django.test import TestCase, override_settings
from django.conf import settings
from django.core import management
from django.urls import reverse
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.auth.hashers import make_password
import rethinkdb as r

from bonk.serializers import *

@override_settings(
    RETHINK_DB_DB=os.environ.get('RETHINK_DB_DB', 'bonkci'),
)
class APITests(TestCase):
    @classmethod
    def setUpClass(cls):
        super(APITests, cls).setUpClass()
        cls.conn = r.connect(host=settings.RETHINK_DB_HOST, port=settings.RETHINK_DB_PORT)
        try:
            r.db_drop(settings.RETHINK_DB_DB).run(cls.conn)
        except:
            pass
        r.db_create(settings.RETHINK_DB_DB).run(cls.conn)
        cls.conn.db = settings.RETHINK_DB_DB
        management.call_command('syncrethinkdb', verbosity=0)

    @classmethod
    def tearDownClass(cls):
        r.db_drop(settings.RETHINK_DB_DB).run(cls.conn)
        super(APITests, cls).tearDownClass()

    def tearDown(self):
        for t in ["vrf", "ip_prefix", "ip_block", "ip_address", "dns_zone", "dns_record", "dhcp_server_set"]:
            r.table(t).delete().run(self.conn)
        super(APITests, self).tearDown()

    def create_user(self, username='tester', password='tester', is_superuser=True, groups=[], **kwargs):
        user = get_user_model().objects.create(
            username=username,
            password=make_password(password),
            is_superuser=is_superuser,
            **kwargs
        )
        for name in groups:
            group, created = Group.objects.get_or_create(name=name)
            user.groups.add(group)
        auth = "Basic %s" % (base64.b64encode("%s:%s" % (username, password)))
        return auth

    def create_common_objects(self):
        auth = self.create_user()
        response = self.client.post(reverse('bonk:vrf_list'), data=json.dumps({
            'vrf': 0, 'name': 'default'
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)
        return auth

    def create_ip_block(self, auth, vrf, network, length, **fields):
        response = self.client.post(reverse('bonk:block_list'), data=json.dumps(dict(fields,
            vrf=vrf,
            network=network,
            length=length,
        )), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def create_ip_prefix(self, auth, vrf, network, length, name, **fields):
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps(dict(fields,
            vrf=vrf,
            network=network,
            length=length,
            name=name,
            state=fields.get('state', 'allocated'),
        )), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def create_ip_address(self, auth, vrf, ip, name, **fields):
        response = self.client.post(reverse('bonk:address_list'), data=json.dumps(dict(fields,
            vrf=vrf,
            ip=ip,
            name=name,
            state=fields.get('state', 'allocated'),
        )), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def _create_zone(self, auth, name, **fields):
        return self.client.post(reverse('bonk:zone_list'), data=json.dumps(dict(fields,
            name=name,
            type=fields.get('type', 'internal'),
        )), content_type="application/json", HTTP_AUTHORIZATION=auth)

    def create_zone(self, auth, name, **fields):
        response = self._create_zone(auth, name, **fields)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def _create_record(self, auth, name, zone, type, value, **fields):
        return self.client.post(reverse('bonk:record_list'), data=json.dumps(dict(fields,
            name=name,
            zone=zone,
            type=type,
            value=value,
        )), content_type="application/json", HTTP_AUTHORIZATION=auth)

    def create_record(self, *args, **fields):
        response = self._create_record(*args, **fields)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def _allocate_ip_prefix(self, auth, vrf, block_network, block_length, **fields):
        return self.client.post(reverse('bonk:block_allocate', kwargs={
                    'vrf': vrf,
                    'network': block_network,
                    'length': block_length,
                }), data=json.dumps(dict(fields,
                    state=fields.get('state', 'allocated'),
                )), content_type="application/json", HTTP_AUTHORIZATION=auth)

    def allocate_ip_prefix(self, *args, **fields):
        response = self._allocate_ip_prefix(*args, **fields)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def _allocate_ip_address(self, auth, vrf, prefix_network, prefix_length, name, **fields):
        return self.client.post(reverse('bonk:prefix_allocate', kwargs={
                    'vrf': vrf,
                    'network': prefix_network,
                    'length': prefix_length
                }), data=json.dumps(dict(fields,
                    name=name,
                    state=fields.get('state', 'allocated'),
                )), content_type="application/json", HTTP_AUTHORIZATION=auth)

    def allocate_ip_address(self, *args, **fields):
        response = self._allocate_ip_address(*args, **fields)
        self.assertEqual(response.status_code, 201)
        return json.loads(response.content)

    def test_ip_block_get_by_ip(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16)
        self.assertEqual(IPBlockSerializer.get_by_ip(0, '10.0.0.0')['id'], ip_block['id'])
        self.assertEqual(IPBlockSerializer.get_by_ip(0, '10.0.255.255')['id'], ip_block['id'])
        with self.assertRaises(RethinkObjectNotFound):
            IPBlockSerializer.get_by_ip(0, '10.1.0.0')

    def test_ip_block_invalid_vrf(self):
        auth = self.create_common_objects()
        response = self.client.post(reverse('bonk:block_list'), data=json.dumps({
            'vrf': 1,
            'network': '10.0.0.0',
            'length': 16,
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('vrf', json.loads(response.content))

    def test_ip_block_invalid_network(self):
        auth = self.create_common_objects()
        response = self.client.post(reverse('bonk:block_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.1.0',
            'length': 16,
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('non_field_errors', json.loads(response.content))

    def test_ip_prefix_invalid_vrf(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16)
        response = self._allocate_ip_prefix(auth, 1, '10.0.0.0', 16, length=24, name='prefix1', permissions={})
        self.assertEqual(response.status_code, 404)

    def test_ip_prefix_get_by_ip(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16)
        ip_prefix = self.create_ip_prefix(auth, 0, '10.0.1.0', 24, 'prefix1')
        self.assertEqual(IPPrefixSerializer.get_by_ip(0, '10.0.1.0')['id'], ip_prefix['id'])
        self.assertEqual(IPPrefixSerializer.get_by_ip(0, '10.0.1.255')['id'], ip_prefix['id'])
        with self.assertRaises(RethinkObjectNotFound):
            IPPrefixSerializer.get_by_ip(0, '10.0.0.0')
        with self.assertRaises(RethinkObjectNotFound):
            IPPrefixSerializer.get_by_ip(0, '10.0.2.0')

    def test_ip_prefix_list_as_user(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1', 'group2']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        ip_prefix2 = self.allocate_ip_prefix(user2_auth, 0, '10.0.0.0', 16, length=24, name='prefix2', permissions={'write': ['group2']})

        response = self.client.get(reverse('bonk:prefix_list'), HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], ip_prefix1['id'])

        response = self.client.get(reverse('bonk:prefix_list'), HTTP_AUTHORIZATION=user2_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], ip_prefix2['id'])

    def test_ip_prefix_allocate_forbidden(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={})
        response = self._allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        self.assertEqual(response.status_code, 403)

    def test_ip_prefix_allocate_hosts(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, hosts=13, name='prefix1', permissions={'write': ['group1']})
        self.assertEqual(ip_prefix1['length'], 28)

    def test_ip_prefix_allocate_nothing(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        response = self._allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, name='prefix1', permissions={'write': ['group1']})
        self.assertEqual(response.status_code, 400)
        self.assertIn('length', response.content)

    def test_ip_prefix_allocate_exhaustive(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=17, name='prefix1', permissions={'write': ['group1']})
        ip_prefix2 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=17, name='prefix2', permissions={'write': ['group1']})
        response = self._allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=17, name='prefix3', permissions={'write': ['group1']})
        self.assertEqual(response.status_code, 400)
        self.assertIn('exhausted', response.content)

    def test_ip_prefix_allocate_no_permissions(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        response = self._allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=17, name='prefix1')
        self.assertEqual(response.status_code, 400)
        self.assertIn('permissions', response.content)

    def test_ip_prefix_no_block(self):
        auth = self.create_common_objects()
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.0',
            'length': 24,
            'state': 'allocated',
            'name': 'prefix1',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('non_field_errors', json.loads(response.content))

    def test_ip_prefix_larger_than_block(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={})
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.0',
            'length': 8,
            'state': 'allocated',
            'name': 'prefix1',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('non_field_errors', json.loads(response.content))

    def test_ip_prefix_overlap(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, allocators=[])
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.0',
            'length': 24,
            'state': 'allocated',
            'name': 'prefix1',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.128',
            'length': 28,
            'state': 'allocated',
            'name': 'prefix2',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('non_field_errors', json.loads(response.content))

    def test_ip_prefix_underlap(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, allocators=[])
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.128',
            'length': 28,
            'state': 'allocated',
            'name': 'prefix1',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.0',
            'length': 24,
            'state': 'allocated',
            'name': 'prefix2',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('non_field_errors', json.loads(response.content))

    def test_ip_prefix_invalid_network(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, allocators=[])
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.0.128',
            'length': 24,
            'state': 'allocated',
            'name': 'prefix1',
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 400)
        self.assertIn('non_field_errors', json.loads(response.content))

    def test_ip_prefix_high_ip(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '128.0.0.0', 24)
        ip_prefix1 = self.allocate_ip_prefix(auth, 0, '128.0.0.0', 24, length=28, name='prefix1', permissions={})

    def test_create_prefix_without_permission(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={})
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        response = self.client.post(reverse('bonk:prefix_list'), data=json.dumps({
            'vrf': 0,
            'network': '10.0.1.0',
            'length': 24,
            'state': 'allocated',
            'name': 'prefix1',
        }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('non_field_errors', data)
        self.assertIn('permission', data['non_field_errors'][0])

    def test_ip_address_allocate(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1', 'group2']})
        zone = self.create_zone(auth, 'my.zone', permissions={'write': ['group1', 'group2']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        ip_prefix2 = self.allocate_ip_prefix(user2_auth, 0, '10.0.0.0', 16, length=24, name='prefix2', permissions={'write': ['group2']})

        ip1 = self.allocate_ip_address(user1_auth, 0, ip_prefix1['network'], ip_prefix1['length'], 'test1.my.zone')
        ip2 = self.allocate_ip_address(user2_auth, 0, ip_prefix2['network'], ip_prefix2['length'], 'test2.my.zone')
        self.assertIn(netaddr.IPAddress(ip1['ip']), netaddr.IPNetwork("%s/%d" % (ip_prefix1['network'], ip_prefix1['length'])))
        self.assertIn(netaddr.IPAddress(ip2['ip']), netaddr.IPNetwork("%s/%d" % (ip_prefix2['network'], ip_prefix2['length'])))

        response = self.client.get(reverse('bonk:address_list'), HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], ip1['id'])
        response = self.client.get(reverse('bonk:address_list'), HTTP_AUTHORIZATION=user2_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['id'], ip2['id'])

    def test_ip_address_allocate_no_zone(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        response = self._allocate_ip_address(user1_auth, 0, ip_prefix1['network'], ip_prefix1['length'], 'test1.my.zone')
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('name', data)
        self.assertIn('matching', data['name'][0])

    def test_ip_address_allocate_no_zone_permission(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={})
        response = self._allocate_ip_address(user1_auth, 0, ip_prefix1['network'], ip_prefix1['length'], 'test1.my.zone')
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('name', data)
        self.assertIn('permission', data['name'][0])

    def test_ip_address_allocate_duplicate_name(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={'write': ['group1']})
        ip1 = self.allocate_ip_address(user1_auth, 0, ip_prefix1['network'], ip_prefix1['length'], 'test1.my.zone')
        response = self._allocate_ip_address(user1_auth, 0, ip_prefix1['network'], ip_prefix1['length'], 'test1.my.zone')
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('name', data)
        self.assertIn('already', data['name'][0])

    def test_ip_address_create_no_prefix(self):
        auth = self.create_common_objects()
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={})
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        zone = self.create_zone(auth, 'my.zone', permissions={'create': ['group1']})
        response = self.client.post(reverse('bonk:address_list'), data=json.dumps({
            'vrf': 0,
            'ip': '10.0.0.2',
            'name': 'test1.my.zone',
            'state': 'allocated',
        }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('non_field_errors', data)
        self.assertIn('no prefix found', data['non_field_errors'][0])

    def test_ip_address_allocate_no_name(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        response = self.client.post(reverse('bonk:prefix_allocate', kwargs={
                'vrf': 0,
                'network': ip_prefix1['network'],
                'length': ip_prefix1['length']
            }), data=json.dumps({
                'vrf': 0,
                'ip': '10.0.0.2',
                'state': 'allocated',
            }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('name', data[0])

    def test_ip_address_allocate_exhaustive(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={'write': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=28, name='prefix1', permissions={'write': ['group1']})
        for i in range(0, 14):
            self.allocate_ip_address(user1_auth, ip_prefix1['vrf'], ip_prefix1['network'], ip_prefix1['length'], "ip%d.my.zone" % i)
        response = self._allocate_ip_address(user1_auth, ip_prefix1['vrf'], ip_prefix1['network'], ip_prefix1['length'], "ip-fail.my.zone")
        self.assertEqual(response.status_code, 400)
        self.assertIn('exhausted', response.content)

    def test_ip_address_allocate_specific(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={'write': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        self.allocate_ip_address(user1_auth, ip_prefix1['vrf'], ip_prefix1['network'], ip_prefix1['length'], "ip2.my.zone", ip='10.0.0.2')
        response = self._allocate_ip_address(user1_auth, ip_prefix1['vrf'], ip_prefix1['network'], ip_prefix1['length'], "ip2.my.zone", ip='10.0.0.2')
        self.assertEqual(response.status_code, 400)
        self.assertIn('already in use', response.content)

    def test_ip_address_allocate_ttl(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={'write': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        response = self.client.post(reverse('bonk:prefix_allocate', kwargs={
                'vrf': 0,
                'network': ip_prefix1['network'],
                'length': ip_prefix1['length']
            }), data=json.dumps({
                'vrf': 0,
                'ip': '10.0.0.2',
                'name': 'test1.my.zone',
                'state': 'allocated',
                'ttl': 300,
            }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 201)
        data = json.loads(response.content)
        self.assertEqual(data['ttl'], 300)

    def test_ip_address_detail(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        ip_block = self.create_ip_block(auth, 0, '10.0.0.0', 16, permissions={'create': ['group1']})
        ip_prefix1 = self.allocate_ip_prefix(user1_auth, 0, '10.0.0.0', 16, length=24, name='prefix1', permissions={'write': ['group1']})
        zone = self.create_zone(auth, 'my.zone', permissions={'write': ['group1']})
        ip1 = self.allocate_ip_address(user1_auth, 0, ip_prefix1['network'], ip_prefix1['length'], 'test1.my.zone')
        for iter_auth, code in [(user2_auth, 403), (user1_auth, 200)]:
            response = self.client.patch(reverse('bonk:address_detail', kwargs={
                    'vrf': ip1['vrf'],
                    'ip': ip1['ip'],
                }), data=json.dumps({
                    'version': ip1['version'],
                    'dhcp_mac': ['de:ad:be:ef:00:01'],
                }), content_type="application/json", HTTP_AUTHORIZATION=iter_auth)
            self.assertEqual(response.status_code, code)

    def test_dns_zones_list(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        zone2 = self.create_zone(auth, 'my2.zone', permissions={'write': ['group2']})
        zone3 = self.create_zone(auth, 'my3.zone', permissions={'write': ['group1', 'group2']})

        response = self.client.get(reverse('bonk:zone_list'), HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 2)
        self.assertEqual(set(map(lambda x: x['id'], data)), set([zone1['id'], zone3['id']]))

        response = self.client.get(reverse('bonk:zone_list'), HTTP_AUTHORIZATION=user2_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 2)
        self.assertEqual(set(map(lambda x: x['id'], data)), set([zone2['id'], zone3['id']]))

    def test_dns_zone_create_without_permission(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        response = self._create_zone(user1_auth, 'my1.zone', permissions={'write': ['group1']})
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('non_field_errors', data)
        self.assertIn('permission', data['non_field_errors'][0])

    def test_dns_zone_create_with_permission(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        root_zone = self.create_zone(auth, 'zone', permissions={'create': ['group1']})
        my_zone = self.create_zone(user1_auth, 'my1.zone', permissions={'write': ['group1']})

    def test_dns_records_list(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        zone2 = self.create_zone(auth, 'my2.zone', permissions={'write': ['group2']})
        record_apex1 = self.create_record(user1_auth, 'my1.zone', 'my1.zone', 'A', ['127.0.0.1'])
        record_www1 = self.create_record(user1_auth, 'www.my1.zone', 'my1.zone', 'A', ['127.0.0.1'])
        record_apex2 = self.create_record(user2_auth, 'my2.zone', 'my2.zone', 'A', ['127.0.0.1'])
        record_www2 = self.create_record(user2_auth, 'www.my2.zone', 'my2.zone', 'A', ['127.0.0.1'])

        response = self.client.get(reverse('bonk:record_list'), HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 2)
        self.assertEqual(set(map(lambda x: x['id'], data)), set([record_apex1['id'], record_www1['id']]))

        response = self.client.get(reverse('bonk:record_list'), HTTP_AUTHORIZATION=user2_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 2)
        self.assertEqual(set(map(lambda x: x['id'], data)), set([record_apex2['id'], record_www2['id']]))

    def test_dns_records_no_zone(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        response = self._create_record(user1_auth, 'my1.zone', 'my1.zone', 'A', ['127.0.0.1'])
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('zone', data)

    def test_dns_records_no_manager(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        response = self._create_record(user2_auth, 'my1.zone', 'my1.zone', 'A', ['127.0.0.1'])
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('zone', data)

    def test_dns_records_name_not_in_zone(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        response = self._create_record(user1_auth, 'my2.zone', 'my1.zone', 'A', ['127.0.0.1'])
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('non_field_errors', data)

    def test_dns_records_cname_for_existing(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        self.create_record(user1_auth, 'service.my1.zone', 'my1.zone', 'A', ['127.0.0.1'])
        response = self._create_record(user1_auth, 'service.my1.zone', 'my1.zone', 'CNAME', ['service2.my2.zone'])
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('non_field_errors', data)

    def test_dns_records_a_for_cname(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        self.create_record(user1_auth, 'service.my1.zone', 'my1.zone', 'CNAME', ['service.my2.zone'])
        response = self._create_record(user1_auth, 'service.my1.zone', 'my1.zone', 'A', ['127.0.0.1'])
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertIn('non_field_errors', data)

    def test_dns_record_detail(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1']})
        record = self.create_record(user1_auth, 'service.my1.zone', 'my1.zone', 'CNAME', ['service.my2.zone'])
        response = self.client.patch(reverse('bonk:record_detail', kwargs={
                'name': record['name'],
                'type': record['type'],
            }), data=json.dumps({
                'version': record['version'],
                'type': 'A',
                'value': ['127.0.0.1'],
            }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)

    def test_dns_records_reviews(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])
        user2_auth = self.create_user('user2', is_superuser=False, groups=['group2'])
        zone1 = self.create_zone(auth, 'my1.zone', permissions={'write': ['group1'], 'create': ['group2']}, needs_review=True)

        response = self._create_record(user2_auth, 'www.my1.zone', 'my1.zone', 'A', ['127.0.0.1'], permissions={'write': ['group2']})
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.content)
        self.assertEqual(data[0], 'review created')

        response = self.client.patch(reverse('django_rethink:review_detail', kwargs={'id': data[1]}), data=json.dumps({
            'approvals': ['user1'],
        }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)

        response = self.client.patch(reverse('django_rethink:review_detail', kwargs={'id': data[1]}), data=json.dumps({
            'state': 'executed',
        }), content_type="application/json", HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(reverse('bonk:record_list'), HTTP_AUTHORIZATION=user2_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['name'], 'www.my1.zone')

    def test_dhcp_server_set_list(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])

        response = self.client.post(reverse('bonk:dhcp_server_set_list'), data=json.dumps({
            'name': 'dhcp-set-1',
            'servers': ['10.0.0.2', '10.0.0.3'],
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)

        response = self.client.get(reverse('bonk:dhcp_server_set_list'), HTTP_AUTHORIZATION=user1_auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)

        response = self.client.get(reverse('bonk:dhcp_server_set_list'), HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertEqual(len(data), 1)

    def test_dhcp_server_set_detail(self):
        auth = self.create_common_objects()
        user1_auth = self.create_user('user1', is_superuser=False, groups=['group1'])

        response = self.client.post(reverse('bonk:dhcp_server_set_list'), data=json.dumps({
            'name': 'dhcp-set-1',
            'servers': ['10.0.0.2', '10.0.0.3'],
        }), content_type="application/json", HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 201)

        response = self.client.get(reverse('bonk:dhcp_server_set_detail', kwargs={'slug': 'dhcp-set-1'}), HTTP_AUTHORIZATION=auth)
        self.assertEqual(response.status_code, 200)



import requests
import subprocess
import logging
import urlparse
from celery import shared_task
from django.conf import settings

logger = logging.getLogger("bonk.tasks")

@shared_task
def trigger_dns_dhcp_rebuild(obj):
    if hasattr(settings, 'BONK_TRIGGER_REBUILD'):
        if settings.BONK_TRIGGER_REBUILD['type'] == 'requests':
            response = getattr(requests, settings.BONK_TRIGGER_REBUILD.get('method', 'get'))( 
                settings.BONK_TRIGGER_REBUILD['uri']
            )
            if response.status_code not in (200, 201, 204):
                logger.error("failed to trigger rebuild: status = %d, text = %r", response.status_code, response.text)
        elif settings.BONK_TRIGGER_REBUILD['type'] == 'subprocess':
            p = subprocess.Popen(settings.BONK_TRIGGER_REBUILD['command'],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            stdout, stderr = p.communicate()
            logger.info("Triggered rebuild")
            for line in stdout.splitlines():
                logger.info(line)
            for line in stderr.splitlines():
                logger.warning(line)
            if p.returncode != 0:
                raise Exception("failed to trigger rebuild: %d", p.returncode)

def socrates_request(method, url, **kwargs):
    return getattr(requests, method.lower())(
        url,
        auth=settings.BONK_SOCRATES_AUTH,
        **kwargs)

@shared_task
def trigger_prefix_create(prefix, block):
    url = urlparse.urlparse(block['announced_by'])
    if url.scheme == 'socrates' and prefix['state'] == 'allocated':
        domain = url.path.strip('/')
        response = socrates_request("get", "https://%s/asset/" % url.netloc,
            params={
                'network__device': domain
            }
        )
        if response.status_code != 200:
            raise Exception("unable to find firewall %s in Socrates: %r" %
                (domain, response.json()))
        firewalls = response.json()

        switch_domains = set()
        # Add switches which are directly connected to firewalls
        for firewall in firewalls:
            for nic in firewall.get('nics', []):
                if 'remote' in nic and 'domain' in nic['remote']:
                    switch_domains.add(nic['remote']['domain'])

        # Find and add switches with the same configuration endpoint
        switch_urls = set()
        for switch_domain in switch_domains:
            response = socrates_request("get",
                "https://%s/asset/?switch__domain=%s" % (url.netloc, switch_domain))
            if response.status_code != 200:
                raise Exception("unable to find switch for domain %s in Socrates: %r" %
                    (switch_domain, response.json()))
            for switch_asset in response.json():
                if 'url' in switch_asset:
                    switch_urls.add(switch_asset['url'])

        for switch_url in switch_urls:
            response = socrates_request("get",
                "https://%s/asset/?url=%s" % (url.netloc, switch_url))
            if response.status_code != 200:
                raise Exception("unable to find switch for URL %s in Socrates: %r" %
                    (switch_url, response.json()))
            for switch_asset in response.json():
                if 'switch' in switch_asset and 'domain' in switch_asset['switch']:
                    switch_domains.add(switch_asset['switch']['domain'])
                for nic in switch_asset.get('nics', []):
                    if 'remote' in nic and 'domain' in nic['remote']:
                        switch_domains.add(nic['remote']['domain'])

        # Create network
        data = dict(urlparse.parse_qsl(url.query))
        domains = dict(
            [(domain, {'name': prefix['name'], 'vlan_id': 0, 'data': data})] +
            [(switch_domain, {'name': prefix['name'], 'vlan_id': 0})
                for switch_domain in switch_domains]
        )
        response = socrates_request("post", "https://%s/network/" % url.netloc,
            json={
                'vrf': prefix['vrf'],
                'network': prefix['network'],
                'length': prefix['length'],
                'permissions': prefix.get('permissions', {}),
                'domains': domains,
            }
        )
        if response.status_code != 201:
            raise Exception("unable to create network in Socrates %d: %r" %
                            (response.status_code, response.content))

@shared_task
def trigger_prefix_delete(prefix, block):
    url = urlparse.urlparse(block['announced_by'])
    if url.scheme == 'socrates' and prefix['state'] == 'allocated':
        response = socrates_request("delete", "https://%s/network/%d/%s/%d/" %
            (url.netloc, prefix['vrf'], prefix['network'], prefix['length']))
        if response.status_code not in (204, 404):
            raise Exception("unable to delete network from Socrates %d: %r" %
                            (response.status_code, response.content))



# Copyright 2017 Klarna Bank AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
import netaddr
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status
from rest_framework import generics
from rest_framework import permissions
from django_rethink import RethinkAPIMixin, RethinkSerializerPermission
from django_rethink.tasks import rethinkdb_lock, rethinkdb_unlock
from bonk.serializers import *

logger = logging.getLogger("bonk.views")

class IsAdminForUpdate(permissions.BasePermission):
    def has_permission(self, request, view):
        if request.method in ('POST', 'PUT', 'PATCH', 'DELETE'):
            return request.user.is_superuser
        else:
            return True

class VRFListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = VRFSerializer
    permission_classes = (permissions.IsAuthenticated, IsAdminForUpdate)

class VRFDetailView(RethinkAPIMixin, generics.RetrieveUpdateAPIView):
    serializer_class = VRFSerializer
    permission_classes = (permissions.IsAuthenticated, IsAdminForUpdate)

    def get_slug(self):
        return int(self.kwargs['vrf'])

class IPBlockListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = IPBlockSerializer
    group_filter_fields = ['permissions_read', 'permissions_write', 'permissions_create']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

class IPBlockDetailView(RethinkAPIMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = IPBlockSerializer
    group_filter_fields = ['permissions_read', 'permissions_write', 'permissions_create']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

    def get_slug(self):
        return [int(self.kwargs['vrf']), self.kwargs['network'], int(self.kwargs['length'])]

class IPBlockAllocateView(RethinkAPIMixin, generics.CreateAPIView):
    serializer_class = IPBlockSerializer
    group_filter_fields = ['permissions_create']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

    def get_slug(self):
        return [int(self.kwargs['vrf']), self.kwargs['network'], int(self.kwargs['length'])]

    def create(self, *args, **kwargs):
        block = self.get_object()
        if 'hosts' in self.request.data and 'length' not in self.request.data:
            if isinstance(self.request.data['hosts'], int):
                s = bin(self.request.data['hosts'] + 2)[2:]
                length = ((32 - len(s)) * "0" + s).find("1")
            else:
                raise serializers.ValidationError("hosts is invalid")
        elif 'length' not in self.request.data:
            raise serializers.ValidationError("prefix length is required")
        elif (not isinstance(self.request.data['length'], int) or
              self.request.data['length'] < block['length'] or
              self.request.data['length'] > 31):
            raise serializers.ValidationError("prefix length is invalid")
        else:
            length = self.request.data['length']
        if 'name' not in self.request.data:
            raise serializers.ValidationError("name is required")
        if 'permissions' not in self.request.data:
            raise serializers.ValidationError("permissions is required")
        lock_token = str(uuid.uuid4())
        lock_name = "block-allocate-%s-%d" % (block['network'], block['length'])
        if 'dryrun' not in self.request.data:
            result = rethinkdb_lock.apply_async(tuple(), {'name': lock_name, 'token': lock_token, 'timeout': 30})
            result.get()
        try:
            pool = netaddr.IPSet([netaddr.IPNetwork("%s/%d" % (block['network'], block['length']))])
            used = netaddr.IPSet()
            for subblock in IPBlockSerializer.filter_by_block(block):
                if (subblock['vrf'] == block['vrf'] and
                        subblock['network'] == block['network'] and
                        subblock['length'] == block['length']):
                    continue
                if subblock['length'] < block['length']:
                    continue
                used.add(netaddr.IPNetwork("%s/%d" % (subblock['network'], subblock['length'])))
            for prefix in IPPrefixSerializer.filter_by_block(block):
                used.add(netaddr.IPNetwork("%s/%d" % (prefix['network'], prefix['length'])))
            available = pool - used
            larger = None
            for prefix in available.iter_cidrs():
                if prefix.prefixlen == length:
                    break
                elif prefix.prefixlen < length and (larger is None or larger.prefixlen < prefix.prefixlen):
                    larger = prefix
            else:
                if larger is None:
                    raise serializers.ValidationError("IP block is exhausted")
                prefix = larger.subnet(length).next()
            obj = {
                'vrf': block['vrf'],
                'network': str(prefix.network),
                'length': prefix.prefixlen,
                'name': self.request.data['name'],
                'state': self.request.data.get('state', 'allocated'),
                'permissions': self.request.data['permissions'],
                'gateway': str(prefix.network + 1),
            }
            for field in ['reference', 'dhcp', 'ddns', 'asn', 'tags']:
                if field in self.request.data:
                    obj[field] = self.request.data[field]
            serializer = IPPrefixSerializer(None, data=obj, context={'request': self.request})
            serializer.is_valid(raise_exception=True)
            if 'dryrun' in self.request.data:
                return Response(status=status.HTTP_204_NO_CONTENT)
            else:
                return Response(serializer.save(), status=status.HTTP_201_CREATED)
        finally:
            if 'dryrun' not in self.request.data:
                rethinkdb_unlock.apply_async(tuple(), {'name': lock_name, 'token': lock_token})

class IPPrefixListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = IPPrefixSerializer
    group_filter_fields = ['permissions_read', 'permissions_create', 'permissions_write']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

class IPPrefixDetailView(RethinkAPIMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = IPPrefixSerializer
    group_filter_fields = ['permissions_read', 'permissions_create', 'permissions_write']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

    def get_slug(self):
        return [int(self.kwargs['vrf']), self.kwargs['network'], int(self.kwargs['length'])]

class IPPrefixAllocateView(RethinkAPIMixin, generics.CreateAPIView):
    serializer_class = IPPrefixSerializer
    group_filter_fields = ['permissions_create']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

    def get_slug(self):
        return [int(self.kwargs['vrf']), self.kwargs['network'], int(self.kwargs['length'])]

    def create(self, *args, **kwargs):
        prefix = self.get_object()
        if 'name' not in self.request.data:
            raise serializers.ValidationError("name is required")
        lock_token = str(uuid.uuid4())
        lock_name = "prefix-allocate-%s-%d" % (prefix['network'], prefix['length'])
        if 'dryrun' not in self.request.data:
            result = rethinkdb_lock.apply_async(tuple(), {'name': lock_name, 'token': lock_token, 'timeout': 30})
            result.get()
        try:
            network = netaddr.IPNetwork("%s/%d" % (prefix['network'], prefix['length']))
            used = netaddr.IPSet()
            for address in IPAddressSerializer.filter_by_prefix(prefix):
                used.add(netaddr.IPAddress(address['ip']))
            if prefix['length'] <= 30:
                used.add(network.network)
                used.add(network.broadcast)

            previous = None
            if 'id' in self.request.data:
                previous = IPAddressSerializer.get(id=self.request.data['id'])
                used.remove(previous['ip'])

            available = netaddr.IPSet(network) - used
            if 'ip' not in self.request.data:
                for address in available:
                    break
                else:
                    raise serializers.ValidationError("network is exhausted")
            else:
                address = netaddr.IPAddress(self.request.data['ip'])
                if address not in available:
                    raise serializers.ValidationError("ip='%s' is already in use" % self.request.data['ip'])
            obj = {
                'state': self.request.data.get('state', 'allocated'),
                'vrf': prefix['vrf'],
                'ip': str(address),
                'name': self.request.data['name'],
            }
            for field in ['reference', 'permissions', 'dhcp_mac', 'ttl']:
                if field in self.request.data:
                    obj[field] = self.request.data[field]
            serializer = IPAddressSerializer(previous, data=obj, context={'request': self.request})
            serializer.is_valid(raise_exception=True)
            if 'dryrun' in self.request.data:
                return Response(status=status.HTTP_204_NO_CONTENT)
            else:
                return Response(serializer.save(), status=status.HTTP_201_CREATED)
        finally:
            if 'dryrun' not in self.request.data:
                rethinkdb_unlock.apply_async(tuple(), {'name': lock_name, 'token': lock_token})

class HasAddressPermission(RethinkSerializerPermission):
    def has_object_permission(self, request, view, obj):
        if super(HasAddressPermission, self).has_object_permission(request, view, obj):
            return True
        permission = self.get_permission(request, view, obj)
        user_groups = set(request.user.groups.all().values_list('name', flat=True))
        prefix = IPPrefixSerializer.get_by_ip(obj['vrf'], obj['ip'])
        return len(user_groups.intersection(self.get_groups(prefix, permission))) > 0

class IPAddressListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = IPAddressSerializer
    permission_classes = (permissions.IsAuthenticated,)

    def default_filter_queryset(self, queryset):
        if self.request.user.is_superuser:
            return queryset
        groups = self.request.user.groups.all().values_list('name', flat=True)
        ip_addresses = reduce(lambda x, y: x.union(y),
                [IPAddressSerializer.filter(reql=True).get_all(*groups, index=i)
                    for i in ['permissions_read', 'permissions_create', 'permissions_write']
                ]
            ).distinct()
        from_prefixes = reduce(lambda x, y: x.union(y),
                [IPPrefixSerializer.filter(reql=True).get_all(*groups, index=i)
                    for i in ['permissions_read', 'permissions_create', 'permissions_write']
                ]
            ).distinct().merge(
                lambda p: {"ip_prefix": r.ip_prefix(p['network'], p['length'])}
            ).inner_join(
                r.table("ip_address").merge(
                    lambda a: {"ip_address": r.ip_address(a['ip'])}
                ),
                lambda p, a: r.ip_prefix_contains(p['ip_prefix'], a['ip_address'])
            ).without("left").merge({"left": {}}).zip()
        return from_prefixes.union(ip_addresses).distinct()

class IPAddressDetailView(RethinkAPIMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = IPAddressSerializer
    permission_classes = (permissions.IsAuthenticated, HasAddressPermission)

    def get_slug(self):
        return [int(self.kwargs['vrf']), self.kwargs['ip']]

class DNSZoneListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = DNSZoneSerializer
    group_filter_fields = ['permissions_read', 'permissions_create', 'permissions_write']
    permission_classes = (permissions.IsAuthenticated,)

class DNSZoneDetailView(RethinkAPIMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = DNSZoneSerializer
    group_filter_fields = ['permissions_read', 'permissions_create', 'permissions_write']
    permission_classes = (permissions.IsAuthenticated, RethinkSerializerPermission)

class DNSRecordListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = DNSRecordSerializer
    permission_classes = (permissions.IsAuthenticated,)

    def default_filter_queryset(self, queryset):
        if self.request.user.is_superuser:
            return queryset
        queryset = queryset. \
            merge(lambda record: {"zone_obj":
                DNSZoneSerializer.filter(name=record['zone'], reql=True).nth(0)
            })
        groups = self.request.user.groups.all().values_list('name', flat=True)
        queryset = queryset.filter(lambda record:
            record['zone_obj']['permissions']['read'].default([]).set_union(
                record['zone_obj']['permissions']['write'].default([])
            ).set_union(
                record['permissions']['read'].default([])
            ).set_union(
                record['permissions']['write'].default([])
            ).set_intersection(groups).count() > 0)
        return queryset

class HasRecordPermission(RethinkSerializerPermission):
    def has_object_permission(self, request, view, obj):
        if super(HasRecordPermission, self).has_object_permission(request, view, obj):
            return True
        permission = self.get_permission(request, view, obj)
        user_groups = set(request.user.groups.all().values_list('name', flat=True))
        zone = DNSZoneSerializer.get(name=obj['zone'])
        return len(user_groups.intersection(self.get_groups(zone, permission))) > 0

class DNSRecordDetailView(RethinkAPIMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = DNSRecordSerializer
    permission_classes = (permissions.IsAuthenticated, HasRecordPermission)

    def get_slug(self):
        return [self.kwargs.get('name'), self.kwargs.get('type')]

class DHCPServerSetListView(RethinkAPIMixin, generics.ListCreateAPIView):
    serializer_class = DHCPServerSetSerializer
    permission_classes = (permissions.IsAuthenticated, IsAdminForUpdate)

class DHCPServerSetDetailView(RethinkAPIMixin, generics.RetrieveUpdateDestroyAPIView):
    serializer_class = DHCPServerSetSerializer
    permission_classes = (permissions.IsAuthenticated, IsAdminForUpdate)



# Copyright 2017 Klarna Bank AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.conf.urls import include, url
from bonk.views import *

app_name = 'bonk'
urlpatterns = [
    url(r'^vrf/?$', VRFListView.as_view(), name='vrf_list'),
    url(r'^vrf/(?P<vrf>[0-9]+)/?$', VRFListView.as_view(), name='vrf_detail'),
    url(r'^block/?$', IPBlockListView.as_view(), name='block_list'),
    url(r'^block/(?P<vrf>[0-9]+)/(?P<network>[0-9a-f.:]+)/(?P<length>[0-9]+)/?$', IPBlockDetailView.as_view(), name='block_detail'),
    url(r'^block/(?P<vrf>[0-9]+)/(?P<network>[0-9a-f.:]+)/(?P<length>[0-9]+)/allocate/?$', IPBlockAllocateView.as_view(), name='block_allocate'),
    url(r'^prefix/?$', IPPrefixListView.as_view(), name='prefix_list'),
    url(r'^prefix/(?P<vrf>[0-9]+)/(?P<network>[0-9a-f.:]+)/(?P<length>[0-9]+)/?$', IPPrefixDetailView.as_view(), name='prefix_detail'),
    url(r'^prefix/(?P<vrf>[0-9]+)/(?P<network>[0-9a-f.:]+)/(?P<length>[0-9]+)/allocate/?$', IPPrefixAllocateView.as_view(), name='prefix_allocate'),
    url(r'^address/?$', IPAddressListView.as_view(), name='address_list'),
    url(r'^address/(?P<vrf>[0-9]+)/(?P<ip>[0-9a-f.:]+)/?$', IPAddressDetailView.as_view(), name='address_detail'),
    url(r'^zone/?$', DNSZoneListView.as_view(), name='zone_list'),
    url(r'^zone/(?P<slug>[A-Za-z0-9-.]+)/?$', DNSZoneDetailView.as_view(), name='zone_detail'),
    url(r'^record/?$', DNSRecordListView.as_view(), name='record_list'),
    url(r'^record/(?P<name>[A-Za-z0-9-._*]+)/(?P<type>[A-Z]+)/?$', DNSRecordDetailView.as_view(), name='record_detail'),
    url(r'^dhcpserverset/?$', DHCPServerSetListView.as_view(), name='dhcp_server_set_list'),
    url(r'^dhcpserverset/(?P<slug>[A-Za-z0-9-._]+)/?$', DHCPServerSetDetailView.as_view(), name='dhcp_server_set_detail'),
]
